function [net,tr,Ac,El]=trainb(net,Pd,Tl,Ai,Q,TS,VV,TV)
%TRAINB Batch training with weight & bias learning rules.
%
%	Syntax
%	
%	  [net,TR,Ac,El] = trainb(net,Pd,Tl,Ai,Q,TS,VV,TV)
%	  info = trainb(code)
%
%	Description
%
%	  TRAINB is not called directly.  Instead it is called by TRAIN for
%	  network's whose NET.trainFcn property is set to 'trainb'.
%
%	  TRAINB trains a network with weight and bias learning rules
%	  with batch updates. The weights and biases are updated at the end of
%	  an entire pass through the input data.
%
%	  TRAINB(NET,Pd,Tl,Ai,Q,TS,VV) takes these inputs,
%	    NET - Neural network.
%	    Pd  - Delayed inputs.
%	    Tl  - Layer targets.
%	    Ai  - Initial input conditions.
%	    Q   - Batch size.
%	    TS  - Time steps.
%	    VV  - Empty matrix [] or structure of validation vectors.
%	    TV  - Empty matrix [] or structure of test vectors.
%	  and returns,
%	    NET - Trained network.
%	    TR  - Training record of various values over each epoch:
%	          TR.epoch - Epoch number.
%	          TR.perf  - Training performance.
%	          TR.vperf - Validation performance.
%	          TR.tperf - Test performance.
%	    Ac  - Collective layer outputs for last epoch.
%	    El  - Layer errors for last epoch.
%
%	  Training occurs according to the TRAINWB's training parameters,
%	  shown here with their default values:
%	    net.trainParam.epochs  100  Maximum number of epochs to train
%	    net.trainParam.goal      0  Performance goal
%	    net.trainParam.max_fail  5  Maximum validation failures
%	    net.trainParam.show     25  Epochs between displays (NaN for no displays)
%	    net.trainParam.time    inf  Maximum time to train in seconds
%
%	  Dimensions for these variables are:
%	    Pd - NoxNixTS cell array, each element Pd{i,j,ts} is a DijxQ matrix.
%	    Tl - NlxTS cell array, each element P{i,ts} is a VixQ matrix or [].
%		Ai - NlxLD cell array, each element Ai{i,k} is an SixQ matrix.
%	  Where
%	    Ni = net.numInputs
%		Nl = net.numLayers
%		LD = net.numLayerDelays
%	    Ri = net.inputs{i}.size
%	    Si = net.layers{i}.size
%	    Vi = net.targets{i}.size
%	    Dij = Ri * length(net.inputWeights{i,j}.delays)
%
%	  If VV or TV is not [], it must be a structure of vectors:
%	    VV.PD, TV.PD - Validation/test delayed inputs.
%	    VV.Tl, TV.Tl - Validation/test layer targets.
%	    VV.Ai, TV.Ai - Validation/test initial input conditions.
%	    VV.Q,  TV.Q  - Validation/test batch size.
%	    VV.TS, TV.TS - Validation/test time steps.
%	  Validation vectors are used to stop training early if the network
%	  performance on the validation vectors fails to improve or remains
%	  the same for MAX_FAIL epochs in a row.  Test vectors are used as
%	  a further check that the network is generalizing well, but do not
%	  have any effect on training.
%
%	  TRAINB(CODE) returns useful information for each CODE string:
%	    'pnames'    - Names of training parameters.
%	    'pdefaults' - Default training parameters.
%
%	Network Use
%
%	  You can create a standard network that uses TRAINB by calling
%	  NEWLIN.
%
%	  To prepare a custom network to be trained with TRAINB:
%	  1) Set NET.trainFcn to 'trainb'.
%	     (This will set NET.trainParam to TRAINB's default parameters.)
%	  2) Set each NET.inputWeights{i,j}.learnFcn to a learning function.
%	     Set each NET.layerWeights{i,j}.learnFcn to a learning function.
%	     Set each NET.biases{i}.learnFcn to a learning function.
%	     (Weight and bias learning parameters will automatically be
%	     set to default values for the given learning function.)
%
%	  To train the network:
%	  1) Set NET.trainParam properties to desired values.
%	  2) Set weight and bias learning parameters to desired values.
%	  3) Call TRAIN.
%
%	  See NEWLIN for training examples.
%
%	Algorithm
%
%	  Each weight and bias updates according to its learning function
%	  after each epoch (one pass through the entire set of input vectors).
%
%	  Training stops when any of these conditions are met:
%	  1) The maximum number of EPOCHS (repetitions) is reached.
%	  2) Performance has been minimized to the GOAL.
%	  3) The maximum amount of TIME has been exceeded.
%	  4) Validation performance has increase more than MAX_FAIL times
%	     since the last time it decreased (when using validation).
%
%	See also NEWP, NEWLIN, TRAIN.

% Mark Beale, 11-31-97
% Copyright 1992-2000 The MathWorks, Inc.
% $Revision: 1.2 $  $Date: 2000/06/15 04:23:25 $

% FUNCTION INFO
% =============

if isstr(net)
  switch (net)
    case 'pnames',
	  net = {'epochs','goal','max_fail','show','time'};
    case 'pdefaults',
	  trainParam.epochs = 100;
	  trainParam.goal = 0;
	  trainParam.max_fail = 5;
	  trainParam.show = 25;
	  trainParam.time = inf;
	  net = trainParam;
    otherwise,
	  error('Unrecognized code.')
  end
  return
end

% CALCULATION
% ===========

% Parameters
epochs = net.trainParam.epochs;
goal = net.trainParam.goal;
max_fail = net.trainParam.max_fail;
show = net.trainParam.show;
time = net.trainParam.time;

% Parameter Checking
if (~isa(epochs,'double')) | (~isreal(epochs)) | (any(size(epochs)) ~= 1) | ...
  (epochs < 1) | (round(epochs) ~= epochs)
  error('Epochs is not a positive integer.')
end
if (~isa(goal,'double')) | (~isreal(goal)) | (any(size(goal)) ~= 1) | ...
  (goal < 0)
  error('Goal is not zero or a positive real value.')
end
if (~isa(max_fail,'double')) | (~isreal(max_fail)) | (any(size(max_fail)) ~= 1) | ...
  (max_fail < 1) | (round(max_fail) ~= max_fail)
  error('Max_fail is not a positive integer.')
end
if (~isa(show,'double')) | (~isreal(show)) | (any(size(show)) ~= 1) | ...
  (isfinite(show) & ((show < 1) | (round(show) ~= show)))
  error('Show is not ''NaN'' or a positive integer.')
end
if (~isa(time,'double')) | (~isreal(time)) | (any(size(time)) ~= 1) | ...
  (time < 0)
  error('Time is not zero or a positive real value.')
end

% Constants
this = 'TRAINB';
numLayers = net.numLayers;
numInputs = net.numInputs;
numLayerDelays = net.numLayerDelays;
needGradient = net.hint.needGradient;
performFcn = net.performFcn;
dPerformFcn = feval(performFcn,'deriv');
doValidation = ~isempty(VV);
doTest = ~isempty(TV);

% Signals
BP = ones(1,Q);
gIW = cell(numLayers,numInputs,TS);
gLW = cell(numLayers,numLayers,TS);
gB = cell(net.numLayers,1,TS);
gA = cell(net.numLayers,TS);
IWLS = cell(numLayers,numInputs);
LWLS = cell(numLayers,numLayers);
BLS = cell(numLayers,1);

% Initialize
flag_stop=0;
stop = '';
startTime = clock;
X = getx(net);
if (doValidation)
  VV.net = net;
  vperf = calcperf(net,X,VV.Pd,VV.Tl,VV.Ai,VV.Q,VV.TS);
  VV.perf = vperf;
  VV.numFail = 0;
end
tr = newtr(epochs,'perf','vperf','tperf');

% Train
for epoch=0:epochs

  % Performance
  [Ac,N,Zl,Zi,Zb] = calca(net,Pd,Ai,Q,TS);
  El = calce(net,Ac,Tl,TS);
  perf = feval(performFcn,El,X,net.performParam);

  % Training Record
  epochPlus1 = epoch+1;
  tr.perf(epochPlus1) = perf;
  if (doValidation)
    tr.vperf(epochPlus1) = vperf;
  end
  if (doTest)
    tr.tperf(epochPlus1) = calcperf(net,X,TV.Pd,TV.Tl,TV.Ai,TV.Q,TV.TS);
  end
  
  % Stopping Criteria
  currentTime = etime(clock,startTime);
  if (epoch == epochs)
    stop = 'Maximum epoch reached.';
  elseif (perf <= goal)
    stop = 'Performance goal met.';
  elseif (currentTime > time)
    stop = 'Maximum time elapsed.';
  elseif (doValidation) & (VV.numFail > max_fail)
    stop = 'Validation stop.';
  elseif flag_stop
    stop = 'User stop.';
  end
  
  % Progress
  if isfinite(show) & (~rem(epoch,show) | length(stop))
    fprintf(this);
    if isfinite(epochs) fprintf(', Epoch %g/%g',epoch, epochs); end
	if isfinite(time) fprintf(', Time %g%%',currentTime/time/100); end
	if isfinite(goal) fprintf(', %s %g/%g.',upper(net.performFcn),perf,goal); end
	fprintf('\n')
	flag_stop=plotperf(tr,goal,this,epoch);
    if length(stop), fprintf('%s, %s\n\n',this,stop); end
  end
 
  % Stop when criteria indicate its time
  if length(stop)
    if (doValidation)
	  net = VV.net;
	end
    break
  end

  % Gradient
  if (needGradient)
    gE = feval(dPerformFcn,'e',El,net,perf,net.performParam);
    [gB,gIW,gLW,gA] = calcgrad(net,Q,Pd,Zb,Zi,Zl,N,Ac,gE,TS);
  end
  
  % Update with Weight and Bias Learning Functions
  for ts=1:TS
    for i=1:numLayers

      % Update Input Weight Values
      for j=find(net.inputConnect(i,:))
        learnFcn = net.inputWeights{i,j}.learnFcn;
        if length(learnFcn)
          [dw,IWLS{i,j}] = feval(learnFcn,net.IW{i,j}, ...
            Pd{i,j,ts},Zi{i,j},N{i},Ac{i,ts+numLayerDelays},Tl{i,ts},El{i,ts},gIW{i,j,ts},...
            gA{i,ts},net.layers{i}.distances,net.inputWeights{i,j}.learnParam,IWLS{i,j});
          net.IW{i,j} = net.IW{i,j} + dw;
        end
      end
  
      % Update Layer Weight Values
      for j=find(net.layerConnect(i,:))
        learnFcn = net.layerWeights{i,j}.learnFcn;
        if length(learnFcn)
          Ad = cell2mat(Ac(j,ts+numLayerDelays-net.layerWeights{i,j}.delays)');
          [dw,LWLS{i,j}] = feval(learnFcn,net.LW{i,j}, ...
            Ad,Zl{i,j},N{i},Ac{i,ts+numLayerDelays},Tl{i,ts},El{i,ts},gLW{i,j,ts},...
            gA{i,ts},net.layers{i}.distances,net.layerWeights{i,j}.learnParam,LWLS{i,j});
          net.LW{i,j} = net.LW{i,j} + dw;
        end
      end

      % Update Bias Values
      if net.biasConnect(i)
        learnFcn = net.biases{i}.learnFcn;
        if length(learnFcn)
          [db,BLS{i}] = feval(learnFcn,net.b{i}, ...
            BP,Zb{i},N{i},Ac{i,ts+numLayerDelays},Tl{i,ts},El{i,ts},gB{i,ts},...
            gA{i,ts},net.layers{i}.distances,net.biases{i}.learnParam,BLS{i});
          net.b{i} = net.b{i} + db;
        end
      end
    end
  end
  X = getx(net);

  % Validation
  if (doValidation)
    vperf = calcperf(net,X,VV.Pd,VV.Tl,VV.Ai,VV.Q,VV.TS);
	if (vperf < VV.perf)
	  VV.perf = vperf; VV.net = net; VV.numFail = 0;
	elseif (vperf > VV.perf)
      VV.numFail = VV.numFail + 1;
	end
  end
end

% Finish
tr = cliptr(tr,epoch);

